const fs = require('fs-extra');
const url = require('url');
const path = require('path');
const zlib = require('zlib');

const slash = require('slash');
const chalk = require('chalk');
const chokidar = require('chokidar');
const { isString, isUndefined, assign, isArray, isPlainObject, isObject, each } = require('lodash/fp');

const defaultConf = require('./config');
const { CONFIG, HOT } = require('./const');
const { logger, mockArray, mockObject, getModalDescriptor, match } = require('./utils');

let log = () => {};

function intercept(res, filePath) {
  const methods = ['send', 'write', 'json', 'end'];
  let body;
  each(method => {
    const rawFn = res[method].bind(res);
    res[method] = function(content, ...others) {
      if (content) {
        if (isUndefined(body)) {
          body = content;
        } else if (Buffer.isBuffer(body)) {
          body = Buffer.concat([body, content]);
        }
      }
      rawFn(content, ...others);
      if (method === 'end') {
        save(filePath, body, res.get('content-encoding'));
      }
    }.bind(res);
  })(methods);
}

function save(filePath, content, encoded) {
  const message = `./${slash(path.relative(path.resolve(), filePath))}已创建`;
  if (encoded) {
    zlib.unzip(content, (err, buffer) => {
      if (err) {
        console.log(err);
      } else {
        fs.outputFileSync(filePath, buffer);
        log(chalk.green(message));
      }
    });
  } else {
    fs.outputFileSync(filePath, content);
    log(chalk.green(message));
  }
}

function autoResponse({ res, config, code = 200, data }) {
  let formatter;
  if (config.formatter) {
    formatter = config.formatter;
  } else if (config.restful) {
    formatter = data => data;
  } else {
    formatter = (data, code) => ({
      [config.dataField]: data,
      [config.codeField]: code
    });
  }
  const body = JSON.stringify(formatter(data, code));

  res.setHeader('Content-Type', 'application/json;charset=UTF-8');

  if (config.restful) {
    res.status(code).end(body);
  } else {
    res.end(body);
  }
}

module.exports = function(config = CONFIG, hot = HOT) {
  if (isString(config)) {
    if (fs.existsSync(config)) {
      config = require(config);
    } else {
      console.warn('.mockrc不存在，使用默认规则');
      config = {};
    }
  }
  if (isString(hot)) {
    if (fs.existsSync(hot)) {
      const hotPath = hot;
      hot = require(hot);
      chokidar.watch(hotPath).on('change', filePath => {
        log(chalk.green(`${path.relative(path.resolve(), filePath)}有改动，重新加载...`));
        delete require.cache[filePath];
        hot = require(filePath);
        log(chalk.green('重新加载完毕'));
      });
    }
  }
  config = assign(defaultConf)(config);
  log = logger(config);
  return function(req, res, next) {
    // 如果pathname匹配ignores则直接传递
    const { pathname } = url.parse(req.url);
    const { method } = req;
    
    if (!match(pathname, config.patterns)) {
      next();
      return;
    }
    // 在hotrc中指定的接口
    const descriptor = getModalDescriptor(hot, pathname, config.restful ? method : void 0);
    if (descriptor) {
      if (isString(descriptor)) {
        if (!isNaN(descriptor)) {
          autoResponse({ res, config, code: +descriptor });
        } else {
          let data;
          try {
            data = JSON.parse(descriptor);
          } catch (e) {
            data = descriptor;
          }
          autoResponse({ res, config, data });
        }
        log(`${chalk.bgBlue(pathname).padEnd(60)}${chalk.blue('from mock base').padStart(30)}`);
      } else if (isObject(descriptor)) {
        if (isArray(descriptor)) {
          if (descriptor.length > 1) {
            autoResponse({ res, config, data: mockArray(descriptor[0], descriptor[1], descriptor[2]) });
          } else {
            autoResponse({ res, config, data: descriptor[0] });
          }
        } else if (isPlainObject(descriptor)) {
          autoResponse({ res, config, data: mockObject(descriptor) });
        }
        log(`${chalk.bgBlue(pathname).padEnd(60)}${chalk.blue('from mock model').padStart(30)}`);
      }
      return;
    }
    const filePath = path.resolve(`${config.dir}/${req.method.toUpperCase()}/${pathname}.json`);
    if (fs.existsSync(filePath)) {
      // cache文件存在
      if (config.mode === 0) {
        // cache模式使用cache
        log(`${chalk.bgGreen(pathname).padEnd(60)}${chalk.green('from cache').padStart(30)}`);
        res.end(fs.readFileSync(filePath));
      } else {
        log(`${chalk.bgRed(pathname).padEnd(60)}${chalk.red('from server').padStart(30)}`);
        if (config.mode === 2) {
          intercept(res, filePath);
        }
        next();
      }
    } else {
      // cache文件不存在
      log(`${chalk.bgRed(pathname).padEnd(60)}${chalk.red('from server').padStart(30)}`);
      intercept(res, filePath);
      next();
    }
  };
};
